# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
import time
from abc import ABC
from datetime import datetime, timezone
from types import TracebackType
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from httpx import URL, AsyncClient, Response, Timeout

from pontos.errors import PontosError
from pontos.helper import snake_case

SLEEP_TIMEOUT = 30.0  # in seconds
DEFAULT_TIMEOUT = 180.0  # three minutes
DEFAULT_TIMEOUT_CONFIG = Timeout(DEFAULT_TIMEOUT)  # three minutes
RETRY_DELAY = 2.0  # in seconds

Headers = Dict[str, str]
Params = Dict[str, Union[str, int]]
JSON = dict[str, Union[int, str, dict[str, Any]]]

__all__ = (
    "convert_camel_case",
    "format_date",
    "now",
    "NVDApi",
    "NVDResults",
)


def now() -> datetime:
    """
    Return current datetime with UTC timezone applied
    """
    return datetime.now(tz=timezone.utc)


def format_date(date: datetime) -> str:
    """
    Format date matching to NVD api

    Args:
        date: Date to format

    Returns:
        Formatted date as string
    """
    return date.isoformat(timespec="seconds")


def convert_camel_case(dct: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert camel case keys into snake case keys

    Args:
        dct: dict to convert

    Returns:
        A dict with key names converted to snake case
    """
    converted = {}
    for key, value in dct.items():
        converted[snake_case(key)] = value
    return converted


class NoMoreResults(PontosError):
    """
    Raised if the NVD API has no more results to consume
    """


class InvalidState(PontosError):
    """
    Raised if the state of the NVD API is invalid
    """


T = TypeVar("T")

result_iterator_func = Callable[[JSON], Iterator[T]]


class NVDResults(Generic[T], AsyncIterable[T], Awaitable["NVDResults"]):
    """
    A generic object for accessing the results of a NVD API response

    It implements the pagination and will issue requests against the NVD API.
    """

    def __init__(
        self,
        api: "NVDApi",
        params: Params,
        result_func: result_iterator_func,
        *,
        request_results: Optional[int] = None,
        results_per_page: Optional[int] = None,
        start_index: int = 0,
    ) -> None:
        self._api = api
        self._params = params
        self._url: Optional[URL] = None

        self._data: Optional[JSON] = None
        self._it: Optional[Iterator[T]] = None
        self._total_results: Optional[int] = None
        self._downloaded_results: int = 0

        self._start_index = start_index
        self._request_results = request_results
        self._results_per_page = results_per_page

        self._current_index = start_index
        self._current_request_results = request_results
        self._current_results_per_page = results_per_page

        self._result_func = result_func

    async def chunks(self) -> AsyncIterator[Sequence[T]]:
        """
        Return the results in chunks

        The size of the chunks is defined by results_per_page.

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...

                async for results in nvd_results.chunks():
                    for result in results:
                        print(result)
        """
        while True:
            try:
                if self._it:
                    yield list(self._it)
                await self._next_iterator()
            except NoMoreResults:
                return

    async def items(self) -> AsyncIterator[T]:
        """
        Return the results of the NVD API response

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...

                async for result in nvd_results.items():
                    print(result)
        """
        while True:
            try:
                if self._it:
                    for result in self._it:
                        yield result
                await self._next_iterator()
            except NoMoreResults:
                return

    async def json(self) -> Optional[JSON]:
        """
        Return the result from the NVD API request as JSON

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...
                while data := await nvd_results.json():
                    print(data)

        Returns:
            The response data as JSON or None if the response is exhausted.
        """
        try:
            if not self._data:
                await self._next_iterator()

            data = self._data
            self._data = None
            return data
        except NoMoreResults:
            return None

    def __len__(self) -> int:
        """
        Get the number of available result items for a NVD API request

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...
                total_results = len(nvd_results) # None because it hasn't been awaited yet
                json = await nvd_results.json() # request the plain JSON data
                total_results = len(nvd_results) # contains the total number of results now

                nvd_results: NVDResults = ...
                total_results = len(nvd_results) # None because it hasn't been awaited yet
                async for result in nvd_results:
                    print(result)
                total_results = len(nvd_results) # contains the total number of results now

        Returns:
            The total number of available results if the NVDResults has been awaited
        """
        if self._total_results is None:
            raise InvalidState(
                f"{self.__class__.__name__} has not been awaited yet."
            )
        return self._total_results

    def __aiter__(self) -> AsyncIterator[T]:
        """
        Return the results of the NVD API response

        Same as the items() method. @see items()

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...

                async for result in nvd_results:
                    print(result)
        """
        return self.items()

    def __await__(self) -> Generator[Any, None, "NVDResults"]:
        """
        Request the next results from the NVD API

        Examples:
            .. code-block:: python

                nvd_results: NVDResults = ...
                print(len(nvd_results)) # None, because no request has been send yet
                await nvd_results # creates a request to the NVD API
                print(len(nvd_results))

        Returns:
            The response data as JSON or None if the response is exhausted.
        """

        return self._next_iterator().__await__()

    async def _load_next_data(self) -> None:
        if (
            self._current_request_results is not None
            and self._downloaded_results >= self._current_request_results
        ):
            raise NoMoreResults()

        if (
            self._total_results is not None
            and self._current_index >= self._total_results
        ):
            raise NoMoreResults()

        params = self._params
        params["startIndex"] = self._current_index

        if self._current_results_per_page is not None:
            params["resultsPerPage"] = self._current_results_per_page

        response = await self._api._get(params=params)
        response.raise_for_status()

        self._url = response.url
        data: JSON = response.json(object_hook=convert_camel_case)

        self._data = data
        self._current_results_per_page = int(data["results_per_page"])  # type: ignore
        self._total_results = int(data["total_results"])  # type: ignore
        self._current_index += self._current_results_per_page
        self._downloaded_results += self._current_results_per_page

        if not self._current_request_results:
            self._current_request_results = self._total_results

        if (
            self._request_results
            and self._downloaded_results + self._current_results_per_page
            > self._request_results
        ):
            # avoid downloading more results then requested
            self._current_results_per_page = (
                self._request_results - self._downloaded_results
            )

    async def _get_next_iterator(self) -> Iterator[T]:
        await self._load_next_data()
        return self._result_func(self._data)  # type: ignore

    async def _next_iterator(self) -> "NVDResults":
        self._it = await self._get_next_iterator()
        return self

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f'url="{self._url}" '
            f"total_results={self._total_results} "
            f"start_index={self._start_index} "
            f"current_index={self._current_index} "
            f"results_per_page={self._results_per_page}>"
        )


class NVDApi(ABC):
    """
    Abstract base class for querying the NIST NVD API.

    Should be used as an async context manager.
    """

    def __init__(
        self,
        url: str,
        *,
        token: Optional[str] = None,
        timeout: Optional[Timeout] = DEFAULT_TIMEOUT_CONFIG,
        rate_limit: bool = True,
        request_attempts: int = 1,
    ) -> None:
        """
        Create a new instance of the CVE API.

        Args:
            url: The API URL to use.
            token: The API key to use. Using an API key allows to run more
                requests at the same time.
            timeout: Timeout settings for the HTTP requests
            rate_limit: Set to False to ignore rate limits. The public rate
                limit (without an API key) is 5 requests in a rolling 30 second
                window. The rate limit with an API key is 50 requests in a
                rolling 30 second window.
                See https://nvd.nist.gov/developers/start-here#divRateLimits
                Default: True.
            request_attempts: The number of attempts per HTTP request. Defaults to 1.
        """
        self._url = url
        self._token = token
        self._client = AsyncClient(http2=True, timeout=timeout)

        if rate_limit:
            self._rate_limit: Optional[int] = 50 if token else 5
        else:
            self._rate_limit = None

        self._request_count = 0
        self._last_sleep = time.monotonic()

        self._request_attempts = request_attempts

    def _request_headers(self) -> Headers:
        """
        Get the default request headers
        """
        headers = {}

        if self._token:
            headers["apiKey"] = self._token

        return headers

    async def _consider_rate_limit(self) -> None:
        """
        Apply rate limit if necessary
        """
        if not self._rate_limit:
            return

        self._request_count += 1
        if self._request_count > self._rate_limit:
            time_since_last_sleep = time.monotonic() - self._last_sleep

            if time_since_last_sleep < SLEEP_TIMEOUT:
                time_to_sleep = SLEEP_TIMEOUT - time_since_last_sleep
                await asyncio.sleep(time_to_sleep)

            self._last_sleep = time.monotonic()
            self._request_count = 0

    async def _get(
        self,
        *,
        params: Optional[Params] = None,
    ) -> Response:
        """
        A request against the NIST NVD REST API.
        """
        headers = self._request_headers()

        for attempt in range(self._request_attempts):
            if attempt > 0:
                delay = RETRY_DELAY**attempt
                await asyncio.sleep(delay)

            await self._consider_rate_limit()
            response = await self._client.get(
                self._url, headers=headers, params=params
            )
            if not response.is_server_error:
                break

        return response

    async def __aenter__(self) -> "NVDApi":
        # reset rate limit counter
        self._request_count = 0
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return await self._client.__aexit__(  # type: ignore
            exc_type, exc_value, traceback
        )



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

# ruff: noqa:E501

from pontos.github.api.dependabot import GitHubAsyncRESTDependabot
from pontos.github.models.base import SortOrder
from pontos.github.models.dependabot import (
    AlertSort,
    AlertState,
    DependencyScope,
    DismissedReason,
    Severity,
)
from tests import AsyncIteratorMock, aiter, anext
from tests.github.api import GitHubAsyncRESTTestCase, create_response

ALERTS = [
    {
        "number": 2,
        "state": "dismissed",
        "dependency": {
            "package": {"ecosystem": "pip", "name": "django"},
            "manifest_path": "path/to/requirements.txt",
            "scope": "runtime",
        },
        "security_advisory": {
            "ghsa_id": "GHSA-rf4j-j272-fj86",
            "cve_id": "CVE-2018-6188",
            "summary": "Django allows remote attackers to obtain potentially sensitive information by leveraging data exposure from the confirm_login_allowed() method, as demonstrated by discovering whether a user account is inactive",
            "description": "django.contrib.auth.forms.AuthenticationForm in Django 2.0 before 2.0.2, and 1.11.8 and 1.11.9, allows remote attackers to obtain potentially sensitive information by leveraging data exposure from the confirm_login_allowed() method, as demonstrated by discovering whether a user account is inactive.",
            "vulnerabilities": [
                {
                    "package": {"ecosystem": "pip", "name": "django"},
                    "severity": "high",
                    "vulnerable_version_range": ">= 2.0.0, < 2.0.2",
                    "first_patched_version": {"identifier": "2.0.2"},
                },
                {
                    "package": {"ecosystem": "pip", "name": "django"},
                    "severity": "high",
                    "vulnerable_version_range": ">= 1.11.8, < 1.11.10",
                    "first_patched_version": {"identifier": "1.11.10"},
                },
            ],
            "severity": "high",
            "cvss": {
                "vector_string": "CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:N/A:N",
                "score": 7.5,
            },
            "cwes": [
                {
                    "cwe_id": "CWE-200",
                    "name": "Exposure of Sensitive Information to an Unauthorized Actor",
                }
            ],
            "identifiers": [
                {"type": "GHSA", "value": "GHSA-rf4j-j272-fj86"},
                {"type": "CVE", "value": "CVE-2018-6188"},
            ],
            "references": [
                {"url": "https://nvd.nist.gov/vuln/detail/CVE-2018-6188"},
                {"url": "https://github.com/advisories/GHSA-rf4j-j272-fj86"},
                {"url": "https://usn.ubuntu.com/3559-1/"},
                {
                    "url": "https://www.djangoproject.com/weblog/2018/feb/01/security-releases/"
                },
                {"url": "http://www.securitytracker.com/id/1040422"},
            ],
            "published_at": "2018-10-03T21:13:54Z",
            "updated_at": "2022-04-26T18:35:37Z",
            "withdrawn_at": None,
        },
        "security_vulnerability": {
            "package": {"ecosystem": "pip", "name": "django"},
            "severity": "high",
            "vulnerable_version_range": ">= 2.0.0, < 2.0.2",
            "first_patched_version": {"identifier": "2.0.2"},
        },
        "url": "https://api.github.com/repos/octo-org/octo-repo/dependabot/alerts/2",
        "html_url": "https://github.com/octo-org/octo-repo/security/dependabot/2",
        "created_at": "2022-06-15T07:43:03Z",
        "updated_at": "2022-08-23T14:29:47Z",
        "dismissed_at": "2022-08-23T14:29:47Z",
        "dismissed_by": {
            "login": "octocat",
            "id": 1,
            "node_id": "MDQ6VXNlcjE=",
            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
            "gravatar_id": "",
            "url": "https://api.github.com/users/octocat",
            "html_url": "https://github.com/octocat",
            "followers_url": "https://api.github.com/users/octocat/followers",
            "following_url": "https://api.github.com/users/octocat/following{/other_user}",
            "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
            "organizations_url": "https://api.github.com/users/octocat/orgs",
            "repos_url": "https://api.github.com/users/octocat/repos",
            "events_url": "https://api.github.com/users/octocat/events{/privacy}",
            "received_events_url": "https://api.github.com/users/octocat/received_events",
            "type": "User",
            "site_admin": False,
        },
        "dismissed_reason": "tolerable_risk",
        "dismissed_comment": "This alert is accurate but we use a sanitizer.",
        "fixed_at": None,
        "repository": {
            "id": 217723378,
            "node_id": "MDEwOlJlcG9zaXRvcnkyMTc3MjMzNzg=",
            "name": "octo-repo",
            "full_name": "octo-org/octo-repo",
            "owner": {
                "login": "octo-org",
                "id": 6811672,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjY4MTE2NzI=",
                "avatar_url": "https://avatars3.githubusercontent.com/u/6811672?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octo-org",
                "html_url": "https://github.com/octo-org",
                "followers_url": "https://api.github.com/users/octo-org/followers",
                "following_url": "https://api.github.com/users/octo-org/following{/other_user}",
                "gists_url": "https://api.github.com/users/octo-org/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octo-org/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octo-org/subscriptions",
                "organizations_url": "https://api.github.com/users/octo-org/orgs",
                "repos_url": "https://api.github.com/users/octo-org/repos",
                "events_url": "https://api.github.com/users/octo-org/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octo-org/received_events",
                "type": "Organization",
                "site_admin": False,
            },
            "private": True,
            "html_url": "https://github.com/octo-org/octo-repo",
            "description": None,
            "fork": False,
            "url": "https://api.github.com/repos/octo-org/octo-repo",
            "archive_url": "https://api.github.com/repos/octo-org/octo-repo/{archive_format}{/ref}",
            "assignees_url": "https://api.github.com/repos/octo-org/octo-repo/assignees{/user}",
            "blobs_url": "https://api.github.com/repos/octo-org/octo-repo/git/blobs{/sha}",
            "branches_url": "https://api.github.com/repos/octo-org/octo-repo/branches{/branch}",
            "collaborators_url": "https://api.github.com/repos/octo-org/octo-repo/collaborators{/collaborator}",
            "comments_url": "https://api.github.com/repos/octo-org/octo-repo/comments{/number}",
            "commits_url": "https://api.github.com/repos/octo-org/octo-repo/commits{/sha}",
            "compare_url": "https://api.github.com/repos/octo-org/octo-repo/compare/{base}...{head}",
            "contents_url": "https://api.github.com/repos/octo-org/octo-repo/contents/{+path}",
            "contributors_url": "https://api.github.com/repos/octo-org/octo-repo/contributors",
            "deployments_url": "https://api.github.com/repos/octo-org/octo-repo/deployments",
            "downloads_url": "https://api.github.com/repos/octo-org/octo-repo/downloads",
            "events_url": "https://api.github.com/repos/octo-org/octo-repo/events",
            "forks_url": "https://api.github.com/repos/octo-org/octo-repo/forks",
            "git_commits_url": "https://api.github.com/repos/octo-org/octo-repo/git/commits{/sha}",
            "git_refs_url": "https://api.github.com/repos/octo-org/octo-repo/git/refs{/sha}",
            "git_tags_url": "https://api.github.com/repos/octo-org/octo-repo/git/tags{/sha}",
            "hooks_url": "https://api.github.com/repos/octo-org/octo-repo/hooks",
            "issue_comment_url": "https://api.github.com/repos/octo-org/octo-repo/issues/comments{/number}",
            "issue_events_url": "https://api.github.com/repos/octo-org/octo-repo/issues/events{/number}",
            "issues_url": "https://api.github.com/repos/octo-org/octo-repo/issues{/number}",
            "keys_url": "https://api.github.com/repos/octo-org/octo-repo/keys{/key_id}",
            "labels_url": "https://api.github.com/repos/octo-org/octo-repo/labels{/name}",
            "languages_url": "https://api.github.com/repos/octo-org/octo-repo/languages",
            "merges_url": "https://api.github.com/repos/octo-org/octo-repo/merges",
            "milestones_url": "https://api.github.com/repos/octo-org/octo-repo/milestones{/number}",
            "notifications_url": "https://api.github.com/repos/octo-org/octo-repo/notifications{?since,all,participating}",
            "pulls_url": "https://api.github.com/repos/octo-org/octo-repo/pulls{/number}",
            "releases_url": "https://api.github.com/repos/octo-org/octo-repo/releases{/id}",
            "stargazers_url": "https://api.github.com/repos/octo-org/octo-repo/stargazers",
            "statuses_url": "https://api.github.com/repos/octo-org/octo-repo/statuses/{sha}",
            "subscribers_url": "https://api.github.com/repos/octo-org/octo-repo/subscribers",
            "subscription_url": "https://api.github.com/repos/octo-org/octo-repo/subscription",
            "tags_url": "https://api.github.com/repos/octo-org/octo-repo/tags",
            "teams_url": "https://api.github.com/repos/octo-org/octo-repo/teams",
            "trees_url": "https://api.github.com/repos/octo-org/octo-repo/git/trees{/sha}",
        },
    },
    {
        "number": 1,
        "state": "open",
        "dependency": {
            "package": {"ecosystem": "pip", "name": "ansible"},
            "manifest_path": "path/to/requirements.txt",
            "scope": "runtime",
        },
        "security_advisory": {
            "ghsa_id": "GHSA-8f4m-hccc-8qph",
            "cve_id": "CVE-2021-20191",
            "summary": "Insertion of Sensitive Information into Log File in ansible",
            "description": "A flaw was found in ansible. Credentials, such as secrets, are being disclosed in console log by default and not protected by no_log feature when using those modules. An attacker can take advantage of this information to steal those credentials. The highest threat from this vulnerability is to data confidentiality.",
            "vulnerabilities": [
                {
                    "package": {"ecosystem": "pip", "name": "ansible"},
                    "severity": "medium",
                    "vulnerable_version_range": ">= 2.9.0, < 2.9.18",
                    "first_patched_version": {"identifier": "2.9.18"},
                },
                {
                    "package": {"ecosystem": "pip", "name": "ansible"},
                    "severity": "medium",
                    "vulnerable_version_range": "< 2.8.19",
                    "first_patched_version": {"identifier": "2.8.19"},
                },
                {
                    "package": {"ecosystem": "pip", "name": "ansible"},
                    "severity": "medium",
                    "vulnerable_version_range": ">= 2.10.0, < 2.10.7",
                    "first_patched_version": {"identifier": "2.10.7"},
                },
            ],
            "severity": "medium",
            "cvss": {
                "vector_string": "CVSS:3.1/AV:L/AC:L/PR:L/UI:N/S:U/C:H/I:N/A:N",
                "score": 5.5,
            },
            "cwes": [
                {
                    "cwe_id": "CWE-532",
                    "name": "Insertion of Sensitive Information into Log File",
                }
            ],
            "identifiers": [
                {"type": "GHSA", "value": "GHSA-8f4m-hccc-8qph"},
                {"type": "CVE", "value": "CVE-2021-20191"},
            ],
            "references": [
                {"url": "https://nvd.nist.gov/vuln/detail/CVE-2021-20191"},
                {
                    "url": "https://access.redhat.com/security/cve/cve-2021-20191"
                },
                {"url": "https://bugzilla.redhat.com/show_bug.cgi?id=1916813"},
            ],
            "published_at": "2021-06-01T17:38:00Z",
            "updated_at": "2021-08-12T23:06:00Z",
            "withdrawn_at": None,
        },
        "security_vulnerability": {
            "package": {"ecosystem": "pip", "name": "ansible"},
            "severity": "medium",
            "vulnerable_version_range": "< 2.8.19",
            "first_patched_version": {"identifier": "2.8.19"},
        },
        "url": "https://api.github.com/repos/octo-org/hello-world/dependabot/alerts/1",
        "html_url": "https://github.com/octo-org/hello-world/security/dependabot/1",
        "created_at": "2022-06-14T15:21:52Z",
        "updated_at": "2022-06-14T15:21:52Z",
        "dismissed_at": None,
        "dismissed_by": None,
        "dismissed_reason": None,
        "dismissed_comment": None,
        "fixed_at": None,
        "repository": {
            "id": 664700648,
            "node_id": "MDEwOlJlcG9zaXRvcnk2NjQ3MDA2NDg=",
            "name": "hello-world",
            "full_name": "octo-org/hello-world",
            "owner": {
                "login": "octo-org",
                "id": 6811672,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjY4MTE2NzI=",
                "avatar_url": "https://avatars3.githubusercontent.com/u/6811672?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octo-org",
                "html_url": "https://github.com/octo-org",
                "followers_url": "https://api.github.com/users/octo-org/followers",
                "following_url": "https://api.github.com/users/octo-org/following{/other_user}",
                "gists_url": "https://api.github.com/users/octo-org/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octo-org/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octo-org/subscriptions",
                "organizations_url": "https://api.github.com/users/octo-org/orgs",
                "repos_url": "https://api.github.com/users/octo-org/repos",
                "events_url": "https://api.github.com/users/octo-org/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octo-org/received_events",
                "type": "Organization",
                "site_admin": False,
            },
            "private": True,
            "html_url": "https://github.com/octo-org/hello-world",
            "description": None,
            "fork": False,
            "url": "https://api.github.com/repos/octo-org/hello-world",
            "archive_url": "https://api.github.com/repos/octo-org/hello-world/{archive_format}{/ref}",
            "assignees_url": "https://api.github.com/repos/octo-org/hello-world/assignees{/user}",
            "blobs_url": "https://api.github.com/repos/octo-org/hello-world/git/blobs{/sha}",
            "branches_url": "https://api.github.com/repos/octo-org/hello-world/branches{/branch}",
            "collaborators_url": "https://api.github.com/repos/octo-org/hello-world/collaborators{/collaborator}",
            "comments_url": "https://api.github.com/repos/octo-org/hello-world/comments{/number}",
            "commits_url": "https://api.github.com/repos/octo-org/hello-world/commits{/sha}",
            "compare_url": "https://api.github.com/repos/octo-org/hello-world/compare/{base}...{head}",
            "contents_url": "https://api.github.com/repos/octo-org/hello-world/contents/{+path}",
            "contributors_url": "https://api.github.com/repos/octo-org/hello-world/contributors",
            "deployments_url": "https://api.github.com/repos/octo-org/hello-world/deployments",
            "downloads_url": "https://api.github.com/repos/octo-org/hello-world/downloads",
            "events_url": "https://api.github.com/repos/octo-org/hello-world/events",
            "forks_url": "https://api.github.com/repos/octo-org/hello-world/forks",
            "git_commits_url": "https://api.github.com/repos/octo-org/hello-world/git/commits{/sha}",
            "git_refs_url": "https://api.github.com/repos/octo-org/hello-world/git/refs{/sha}",
            "git_tags_url": "https://api.github.com/repos/octo-org/hello-world/git/tags{/sha}",
            "hooks_url": "https://api.github.com/repos/octo-org/hello-world/hooks",
            "issue_comment_url": "https://api.github.com/repos/octo-org/hello-world/issues/comments{/number}",
            "issue_events_url": "https://api.github.com/repos/octo-org/hello-world/issues/events{/number}",
            "issues_url": "https://api.github.com/repos/octo-org/hello-world/issues{/number}",
            "keys_url": "https://api.github.com/repos/octo-org/hello-world/keys{/key_id}",
            "labels_url": "https://api.github.com/repos/octo-org/hello-world/labels{/name}",
            "languages_url": "https://api.github.com/repos/octo-org/hello-world/languages",
            "merges_url": "https://api.github.com/repos/octo-org/hello-world/merges",
            "milestones_url": "https://api.github.com/repos/octo-org/hello-world/milestones{/number}",
            "notifications_url": "https://api.github.com/repos/octo-org/hello-world/notifications{?since,all,participating}",
            "pulls_url": "https://api.github.com/repos/octo-org/hello-world/pulls{/number}",
            "releases_url": "https://api.github.com/repos/octo-org/hello-world/releases{/id}",
            "stargazers_url": "https://api.github.com/repos/octo-org/hello-world/stargazers",
            "statuses_url": "https://api.github.com/repos/octo-org/hello-world/statuses/{sha}",
            "subscribers_url": "https://api.github.com/repos/octo-org/hello-world/subscribers",
            "subscription_url": "https://api.github.com/repos/octo-org/hello-world/subscription",
            "tags_url": "https://api.github.com/repos/octo-org/hello-world/tags",
            "teams_url": "https://api.github.com/repos/octo-org/hello-world/teams",
            "trees_url": "https://api.github.com/repos/octo-org/hello-world/git/trees{/sha}",
        },
    },
]


class GitHubAsyncRESTDependabotTestCase(GitHubAsyncRESTTestCase):
    api_cls = GitHubAsyncRESTDependabot

    async def test_enterprise_alerts(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.enterprise_alerts("foo"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={"per_page": "100", "sort": "created", "direction": "desc"},
        )

    async def test_enterprise_alerts_state(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.enterprise_alerts("foo", state=AlertState.FIXED)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "state": "fixed",
            },
        )

    async def test_enterprise_alerts_severity(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.enterprise_alerts("foo", severity=Severity.CRITICAL)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "severity": "critical",
            },
        )

    async def test_enterprise_alerts_ecosystem(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.enterprise_alerts("foo", ecosystem="pip"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "ecosystem": "pip",
            },
        )

    async def test_enterprise_alerts_scope(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.enterprise_alerts("foo", scope=DependencyScope.RUNTIME)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "scope": "runtime",
            },
        )

    async def test_enterprise_alerts_sort(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.enterprise_alerts("foo", sort=AlertSort.UPDATED)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "updated",
                "direction": "desc",
            },
        )

    async def test_enterprise_alerts_direction(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.enterprise_alerts("foo", direction=SortOrder.ASC)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/enterprises/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "asc",
            },
        )

    async def test_organization_alerts(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.organization_alerts("foo"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={"per_page": "100", "sort": "created", "direction": "desc"},
        )

    async def test_organization_alerts_state(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.organization_alerts("foo", state=AlertState.FIXED)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "state": "fixed",
            },
        )

    async def test_organization_alerts_severity(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.organization_alerts("foo", severity=Severity.CRITICAL)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "severity": "critical",
            },
        )

    async def test_organization_alerts_ecosystem(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.organization_alerts("foo", ecosystem="pip"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "ecosystem": "pip",
            },
        )

    async def test_organization_alerts_scope(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.organization_alerts("foo", scope=DependencyScope.RUNTIME)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "scope": "runtime",
            },
        )

    async def test_organization_alerts_sort(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.organization_alerts("foo", sort=AlertSort.UPDATED)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "updated",
                "direction": "desc",
            },
        )

    async def test_organization_alerts_direction(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.organization_alerts("foo", direction=SortOrder.ASC)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "asc",
            },
        )

    async def test_alerts(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={"per_page": "100", "sort": "created", "direction": "desc"},
        )

    async def test_alerts_state(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar", state=AlertState.FIXED))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "state": "fixed",
            },
        )

    async def test_alerts_severity(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar", severity=Severity.CRITICAL))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "severity": "critical",
            },
        )

    async def test_alerts_ecosystem(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar", ecosystem="pip"))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "ecosystem": "pip",
            },
        )

    async def test_alerts_scope(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(
            self.api.alerts("foo/bar", scope=DependencyScope.RUNTIME)
        )
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "desc",
                "scope": "runtime",
            },
        )

    async def test_alerts_sort(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar", sort=AlertSort.UPDATED))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "updated",
                "direction": "desc",
            },
        )

    async def test_alerts_direction(self):
        response = create_response()
        response.json.return_value = ALERTS

        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.alerts("foo/bar", direction=SortOrder.ASC))
        alert = await anext(async_it)
        self.assertEqual(alert.number, 2)
        alert = await anext(async_it)
        self.assertEqual(alert.number, 1)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/dependabot/alerts",
            params={
                "per_page": "100",
                "sort": "created",
                "direction": "asc",
            },
        )

    async def test_alert(self):
        response = create_response()
        response.json.return_value = {
            "number": 1,
            "state": "open",
            "dependency": {
                "package": {"ecosystem": "pip", "name": "ansible"},
                "manifest_path": "path/to/requirements.txt",
                "scope": "runtime",
            },
            "security_advisory": {
                "ghsa_id": "GHSA-8f4m-hccc-8qph",
                "cve_id": "CVE-2021-20191",
                "summary": "Insertion of Sensitive Information into Log File in ansible",
                "description": "A flaw was found in ansible. Credentials, such as secrets, are being disclosed in console log by default and not protected by no_log feature when using those modules. An attacker can take advantage of this information to steal those credentials. The highest threat from this vulnerability is to data confidentiality.",
                "vulnerabilities": [
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": ">= 2.9.0, < 2.9.18",
                        "first_patched_version": {"identifier": "2.9.18"},
                    },
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": "< 2.8.19",
                        "first_patched_version": {"identifier": "2.8.19"},
                    },
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": ">= 2.10.0, < 2.10.7",
                        "first_patched_version": {"identifier": "2.10.7"},
                    },
                ],
                "severity": "medium",
                "cvss": {
                    "vector_string": "CVSS:3.1/AV:L/AC:L/PR:L/UI:N/S:U/C:H/I:N/A:N",
                    "score": 5.5,
                },
                "cwes": [
                    {
                        "cwe_id": "CWE-532",
                        "name": "Insertion of Sensitive Information into Log File",
                    }
                ],
                "identifiers": [
                    {"type": "GHSA", "value": "GHSA-8f4m-hccc-8qph"},
                    {"type": "CVE", "value": "CVE-2021-20191"},
                ],
                "references": [
                    {"url": "https://nvd.nist.gov/vuln/detail/CVE-2021-20191"},
                    {
                        "url": "https://access.redhat.com/security/cve/cve-2021-20191"
                    },
                    {
                        "url": "https://bugzilla.redhat.com/show_bug.cgi?id=1916813"
                    },
                ],
                "published_at": "2021-06-01T17:38:00Z",
                "updated_at": "2021-08-12T23:06:00Z",
                "withdrawn_at": None,
            },
            "security_vulnerability": {
                "package": {"ecosystem": "pip", "name": "ansible"},
                "severity": "medium",
                "vulnerable_version_range": "< 2.8.19",
                "first_patched_version": {"identifier": "2.8.19"},
            },
            "url": "https://api.github.com/repos/octocat/hello-world/dependabot/alerts/1",
            "html_url": "https://github.com/octocat/hello-world/security/dependabot/1",
            "created_at": "2022-06-14T15:21:52Z",
            "updated_at": "2022-06-14T15:21:52Z",
            "dismissed_at": None,
            "dismissed_by": None,
            "dismissed_reason": None,
            "dismissed_comment": None,
            "fixed_at": None,
        }
        self.client.get.return_value = response

        alert = await self.api.alert(
            "foo/bar",
            1,
        )

        self.client.get.assert_awaited_once_with(
            "/repos/foo/bar/dependabot/alerts/1",
        )

        self.assertEqual(alert.number, 1)

    async def test_update(self):
        response = create_response()
        response.json.return_value = {
            "number": 1,
            "state": "open",
            "dependency": {
                "package": {"ecosystem": "pip", "name": "ansible"},
                "manifest_path": "path/to/requirements.txt",
                "scope": "runtime",
            },
            "security_advisory": {
                "ghsa_id": "GHSA-8f4m-hccc-8qph",
                "cve_id": "CVE-2021-20191",
                "summary": "Insertion of Sensitive Information into Log File in ansible",
                "description": "A flaw was found in ansible. Credentials, such as secrets, are being disclosed in console log by default and not protected by no_log feature when using those modules. An attacker can take advantage of this information to steal those credentials. The highest threat from this vulnerability is to data confidentiality.",
                "vulnerabilities": [
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": ">= 2.9.0, < 2.9.18",
                        "first_patched_version": {"identifier": "2.9.18"},
                    },
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": "< 2.8.19",
                        "first_patched_version": {"identifier": "2.8.19"},
                    },
                    {
                        "package": {"ecosystem": "pip", "name": "ansible"},
                        "severity": "medium",
                        "vulnerable_version_range": ">= 2.10.0, < 2.10.7",
                        "first_patched_version": {"identifier": "2.10.7"},
                    },
                ],
                "severity": "medium",
                "cvss": {
                    "vector_string": "CVSS:3.1/AV:L/AC:L/PR:L/UI:N/S:U/C:H/I:N/A:N",
                    "score": 5.5,
                },
                "cwes": [
                    {
                        "cwe_id": "CWE-532",
                        "name": "Insertion of Sensitive Information into Log File",
                    }
                ],
                "identifiers": [
                    {"type": "GHSA", "value": "GHSA-8f4m-hccc-8qph"},
                    {"type": "CVE", "value": "CVE-2021-20191"},
                ],
                "references": [
                    {"url": "https://nvd.nist.gov/vuln/detail/CVE-2021-20191"},
                    {
                        "url": "https://access.redhat.com/security/cve/cve-2021-20191"
                    },
                    {
                        "url": "https://bugzilla.redhat.com/show_bug.cgi?id=1916813"
                    },
                ],
                "published_at": "2021-06-01T17:38:00Z",
                "updated_at": "2021-08-12T23:06:00Z",
                "withdrawn_at": None,
            },
            "security_vulnerability": {
                "package": {"ecosystem": "pip", "name": "ansible"},
                "severity": "medium",
                "vulnerable_version_range": "< 2.8.19",
                "first_patched_version": {"identifier": "2.8.19"},
            },
            "url": "https://api.github.com/repos/octocat/hello-world/dependabot/alerts/1",
            "html_url": "https://github.com/octocat/hello-world/security/dependabot/1",
            "created_at": "2022-06-14T15:21:52Z",
            "updated_at": "2022-06-14T15:21:52Z",
            "dismissed_at": None,
            "dismissed_by": None,
            "dismissed_reason": None,
            "dismissed_comment": None,
            "fixed_at": None,
        }
        self.client.patch.return_value = response

        alert = await self.api.update_alert(
            "foo/bar",
            1,
            AlertState.DISMISSED,
            dismissed_reason=DismissedReason.NOT_USED,
            dismissed_comment="Dependency is not used.",
        )

        self.client.patch.assert_awaited_once_with(
            "/repos/foo/bar/dependabot/alerts/1",
            data={
                "state": "dismissed",
                "dismissed_reason": "not_used",
                "dismissed_comment": "Dependency is not used.",
            },
        )

        self.assertEqual(alert.number, 1)



# Copyright (C) 2023 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from argparse import ArgumentTypeError

from ._pep440 import PEP440VersioningScheme
from ._scheme import VersioningScheme
from ._semantic import SemanticVersioningScheme

__all__ = (
    "VERSIONING_SCHEMES",
    "versioning_scheme_argument_type",
    "VersioningScheme",
    "PEP440VersioningScheme",
    "SemanticVersioningScheme",
)

#: Dictionary with available versioning schemes
VERSIONING_SCHEMES: dict[str, type[VersioningScheme]] = {
    "pep440": PEP440VersioningScheme,
    "semver": SemanticVersioningScheme,
}


def versioning_scheme_argument_type(value: str) -> type[VersioningScheme]:
    """
    Verifies if the passed value is a valid versioning scheme and returns
    the corresponding versioning scheme.

    Intended to be used as in `ArgumentParser.add_argument` as the type.

    Raises:
        ArgumentTypeError: If the passed value is not a valid versioning scheme

    Example:
        .. code-block:: python

            from argparse import ArgumentParser
            from pontos.version.scheme versioning_scheme_argument_type

            parser = ArgumentParser()
            parser.add_argument(
                "--versioning-scheme",
                type=versioning_scheme_argument_type,
            )
    """
    try:
        return VERSIONING_SCHEMES[value]
    except KeyError:
        raise ArgumentTypeError(
            f"invalid value {value}. Expected one of "
            f"{', '.join(VERSIONING_SCHEMES.keys())}."
        ) from None



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from argparse import ArgumentParser, Namespace
from typing import Optional, Sequence

import shtab


def cves_parser(args: Optional[Sequence[str]] = None) -> Namespace:
    parser = ArgumentParser()
    shtab.add_argument_to(parser)
    parser.add_argument("--token", help="API key to use for querying.")
    parser.add_argument(
        "--keywords",
        nargs="*",
        help="Search for CVEs containing the keyword in their description.",
    )
    parser.add_argument(
        "--cpe-name", help="Get all CVE information associated with the CPE"
    )
    parser.add_argument(
        "--cvss-v2-vector",
        help="Get all CVE information with the CVSSv2 vector",
    )
    parser.add_argument(
        "--cvss-v3-vector",
        help="Get all CVE information with the CVSSv3 vector",
    )
    parser.add_argument(
        "--source-identifier",
        help="Get all CVE information with the source identifier. For example: "
        "cve@mitre.org",
    )
    parser.add_argument(
        "--number", "-n", metavar="N", help="Request only N CVEs", type=int
    )
    parser.add_argument(
        "--start",
        "-s",
        help="Index of the first CVE to request.",
        type=int,
    )
    return parser.parse_args(args)


def cve_parser(args: Optional[Sequence[str]] = None) -> Namespace:
    parser = ArgumentParser()
    shtab.add_argument_to(parser)
    parser.add_argument("--token", help="API key to use for querying.")
    parser.add_argument("cve_id", metavar="CVE-ID", help="ID of the CVE")
    return parser.parse_args(args)



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import unittest

from pontos.version.schemes._pep440 import PEP440Version
from pontos.version.schemes._semantic import SemanticVersion


class VersionTestCase(unittest.TestCase):
    def test_equal(self):
        self.assertEqual(PEP440Version("22.2.2"), PEP440Version("22.2.2"))
        self.assertNotEqual(PEP440Version("22.2.1"), PEP440Version("22.2.2"))
        self.assertNotEqual(SemanticVersion("22.2.2"), "22.2.2")
        self.assertNotEqual(
            SemanticVersion("22.2.1"), SemanticVersion("22.2.2")
        )
        self.assertNotEqual(PEP440Version("22.2.2"), "22.2.2")

    def test_equal_other(self):
        self.assertEqual(PEP440Version("22.2.2"), SemanticVersion("22.2.2"))

    def test_equal_other_other_way_around(self):
        self.assertEqual(SemanticVersion("22.2.2"), PEP440Version("22.2.2"))

    def test_equal_raises(self):
        with self.assertRaises(ValueError):
            self.assertNotEqual(SemanticVersion("22.2.2"), 22)

        with self.assertRaises(ValueError):
            self.assertNotEqual(PEP440Version("22.2.2"), 22)



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#


from datetime import datetime
from types import TracebackType
from typing import (
    Any,
    Iterator,
    List,
    Optional,
    Type,
    Union,
)
from uuid import UUID

from httpx import Timeout

from pontos.errors import PontosError
from pontos.nvd.api import (
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    NVDApi,
    NVDResults,
    Params,
    convert_camel_case,
    format_date,
    now,
)
from pontos.nvd.models.cpe import CPE

DEFAULT_NIST_NVD_CPES_URL = "https://services.nvd.nist.gov/rest/json/cpes/2.0"
MAX_CPES_PER_PAGE = 10000


def _result_iterator(data: JSON) -> Iterator[CPE]:
    results: list[dict[str, Any]] = data.get("products", [])  # type: ignore
    return (CPE.from_dict(result["cpe"]) for result in results)


class CPEApi(NVDApi):
    """
    API for querying the NIST NVD CPE information.

    Should be used as an async context manager.

    Example:
        .. code-block:: python

            from pontos.nvd.cpe import CPEApi

            async with CPEApi() as api:
                cpe = await api.cpe(...)
    """

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        timeout: Optional[Timeout] = DEFAULT_TIMEOUT_CONFIG,
        rate_limit: bool = True,
        request_attempts: int = 1,
    ) -> None:
        """
        Create a new instance of the CPE API.

        Args:
            token: The API key to use. Using an API key allows to run more
                requests at the same time.
            timeout: Timeout settings for the HTTP requests
            rate_limit: Set to False to ignore rate limits. The public rate
                limit (without an API key) is 5 requests in a rolling 30 second
                window. The rate limit with an API key is 50 requests in a
                rolling 30 second window.
                See https://nvd.nist.gov/developers/start-here#divRateLimits
                Default: True.
            request_attempts: The number of attempts per HTTP request. Defaults to 1.
        """
        super().__init__(
            DEFAULT_NIST_NVD_CPES_URL,
            token=token,
            timeout=timeout,
            rate_limit=rate_limit,
            request_attempts=request_attempts,
        )

    async def cpe(self, cpe_name_id: Union[str, UUID]) -> CPE:
        """
        Query for a CPE matching the CPE UUID.

        Args:
            cpe_name_id: Returns a specific CPE record identified by a Universal
                Unique Identifier (UUID).

        Example:
            .. code-block:: python

                from pontos.nvd.cpe import CPEApi

                async with CPEApi() as api:
                    cpe = await api.cpe("87316812-5F2C-4286-94FE-CC98B9EAEF53")
                    print(cpe)

        Returns:
            A single CPE matching the CPE UUID

        Raises:
            PontosError: If a CPE with the CPE UUID couldn't be found.
        """
        if not cpe_name_id:
            raise PontosError("Missing CPE Name ID.")

        response = await self._get(params={"cpeNameId": str(cpe_name_id)})
        response.raise_for_status()
        data = response.json(object_hook=convert_camel_case)
        products = data["products"]
        if not products:
            raise PontosError(f"No CPE with CPE Name ID '{cpe_name_id}' found.")

        product = products[0]
        return CPE.from_dict(product["cpe"])

    def cpes(
        self,
        *,
        last_modified_start_date: Optional[datetime] = None,
        last_modified_end_date: Optional[datetime] = None,
        cpe_match_string: Optional[str] = None,
        keywords: Optional[Union[List[str], str]] = None,
        match_criteria_id: Optional[str] = None,
        request_results: Optional[int] = None,
        start_index: int = 0,
        results_per_page: Optional[int] = None,
    ) -> NVDResults[CPE]:
        """
        Get all CPEs for the provided arguments

        https://nvd.nist.gov/developers/products

        Args:
            last_modified_start_date: Return all CPEs modified after this date.
            last_modified_end_date: Return all CPEs modified before this date.
                If last_modified_start_date is set but no
                last_modified_end_date is passed it is set to now.
            cpe_match_string: Returns all CPE names that exist in the Official
                CPE Dictionary.
            keywords: Returns only the CPEs where a word or phrase is found in
                the metadata title or reference links.
            match_criteria_id: Returns all CPE records associated with a match
                string identified by its UUID.
            request_results: Number of CPEs to download. Set to None (default)
                to download all available CPEs.
            start_index: Index of the first CPE to be returned. Useful only for
                paginated requests that should not start at the first page.
            results_per_page: Number of results in a single requests. Mostly
                useful for paginated requests.

        Returns:
            A NVDResponse for CPEs

        Examples:
            .. code-block:: python

                from pontos.nvd.cpe import CPEApi

                async with CPEApi() as api:
                    async for cpe in api.cpes(keywords=["Mac OS X"]):
                        print(cpe.cpe_name, cpe.cpe_name_id)

                    json = await api.cpes(request_results=10).json()

                    async for cpes in api.cpes(
                        cpe_match_string="cpe:2.3:o:microsoft:windows_7:-:*:*:*:*:*:*:*",
                    ).chunks():
                        for cpe in cpes:
                            print(cpe)
        """
        params: Params = {}
        if last_modified_start_date:
            params["lastModStartDate"] = format_date(last_modified_start_date)
            if not last_modified_end_date:
                params["lastModEndDate"] = format_date(now())
        if last_modified_end_date:
            params["lastModEndDate"] = format_date(last_modified_end_date)

        if cpe_match_string:
            params["cpeMatchString"] = cpe_match_string

        if keywords:
            if isinstance(keywords, str):
                keywords = [keywords]

            params["keywordSearch"] = " ".join(keywords)
            if any((" " in keyword for keyword in keywords)):
                params["keywordExactMatch"] = ""

        if match_criteria_id:
            params["matchCriteriaId"] = match_criteria_id

        results_per_page = min(
            results_per_page or MAX_CPES_PER_PAGE,
            request_results or MAX_CPES_PER_PAGE,
        )

        return NVDResults(
            self,
            params,
            _result_iterator,
            request_results=request_results,
            results_per_page=results_per_page,
            start_index=start_index,
        )

    async def __aenter__(self) -> "CPEApi":
        await super().__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return await super().__aexit__(  # type: ignore
            exc_type, exc_value, traceback
        )



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#


from typing import AsyncIterator, Iterable, Union

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.api.helper import JSON


class GitHubAsyncRESTLabels(GitHubAsyncREST):
    async def get_all(
        self,
        repo: str,
        issue: Union[int, str],
    ) -> AsyncIterator[str]:
        """
        Get all labels that are set in the issue/pr

        Args:
            repo:   GitHub repository (owner/name) to use
            issue:  Issue/Pull request number

        Returns:
            An async iterator yielding the labels

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async for label in api.labels.get_all("foo/bar", 123):
                        print(label)
        """
        api = f"/repos/{repo}/issues/{issue}/labels"
        params = {"per_page": "100"}

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()
            data: JSON = response.json()

            for label in data:
                yield label["name"]  # type: ignore

    async def delete_all(self, repo: str, issue: Union[int, str]) -> None:
        """
        Deletes all labels in the issue/pr.

        Args:
            repo:   GitHub repository (owner/name) to use
            issue:  Issue/Pull request number

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.labels.delete_all("foo/bar", 123)
        """
        api = f"/repos/{repo}/issues/{issue}/labels"
        response = await self._client.delete(api)
        response.raise_for_status()

    async def set_all(
        self, repo: str, issue: Union[int, str], labels: Iterable[str]
    ) -> None:
        """
        Set labels in the issue/pr.

        Args:
            repo:   GitHub repository (owner/name) to use
            issue:  Issue/Pull request number
            labels: Iterable of labels, that should be set. Existing labels will
                be overwritten.

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.labels.set_all("foo/bar", 123, ["bug", "doc"])
        """
        api = f"/repos/{repo}/issues/{issue}/labels"
        data: JSON = {"labels": labels}  # type: ignore
        response = await self._client.post(api, data=data)
        response.raise_for_status()



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later


import unittest
from argparse import ArgumentTypeError

from pontos.enum import StrEnum, enum_type


class EnumTypeTestCase(unittest.TestCase):
    def test_enum_type(self):
        class FooEnum(StrEnum):
            ALL = "all"
            NONE = "none"

        func = enum_type(FooEnum)

        self.assertEqual(func("all"), FooEnum.ALL)
        self.assertEqual(func("none"), FooEnum.NONE)

        self.assertEqual(func(FooEnum.ALL), FooEnum.ALL)
        self.assertEqual(func(FooEnum.NONE), FooEnum.NONE)

    def test_enum_type_error(self):
        class FooEnum(StrEnum):
            ALL = "all"
            NONE = "none"

        func = enum_type(FooEnum)

        with self.assertRaisesRegex(
            ArgumentTypeError,
            r"invalid value foo. Expected one of all, none",
        ):
            func("foo")



# SPDX-FileCopyrightText: 2020-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def use_cwd(path: Path) -> None:
    """
    Context Manager to change the current working directory temporaryly
    """
    current_cwd = Path.cwd()

    os.chdir(str(path))

    yield

    os.chdir(str(current_cwd))



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta
from typing import Any, Iterator, Optional, Union

from httpx import Timeout

from pontos.errors import PontosError
from pontos.nvd.api import (
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    NVDApi,
    NVDResults,
    Params,
    format_date,
    now,
)
from pontos.nvd.models.cve_change import CVEChange, EventName

__all__ = ("CVEChangesApi",)

DEFAULT_NIST_NVD_CVE_HISTORY_URL = (
    "https://services.nvd.nist.gov/rest/json/cvehistory/2.0"
)

MAX_CVE_CHANGES_PER_PAGE = 5000


def _result_iterator(data: JSON) -> Iterator[CVEChange]:
    results: list[dict[str, Any]] = data.get("cve_changes", [])  # type: ignore
    return (CVEChange.from_dict(result["change"]) for result in results)


class CVEChangesApi(NVDApi):
    """
    API for querying the NIST NVD CVE Change History information.

    Should be used as an async context manager.

    Example:
        .. code-block:: python

            from pontos.nvd.cve_changes import CVEChangesApi

            async with CVEChangesApi() as api:
                async for cve_change in api.changes(event_name=EventName.INITIAL_ANALYSIS):
                    print(cve_change)
    """

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        timeout: Optional[Timeout] = DEFAULT_TIMEOUT_CONFIG,
        rate_limit: bool = True,
        request_attempts: int = 1,
    ) -> None:
        """
        Create a new instance of the CVE Change History API.

        Args:
            token: The API key to use. Using an API key allows to run more
                requests at the same time.
            timeout: Timeout settings for the HTTP requests
            rate_limit: Set to False to ignore rate limits. The public rate
                limit (without an API key) is 5 requests in a rolling 30 second
                window. The rate limit with an API key is 50 requests in a
                rolling 30 second window.
                See https://nvd.nist.gov/developers/start-here#divRateLimits
                Default: True.
            request_attempts: The number of attempts per HTTP request. Defaults to 1.
        """
        super().__init__(
            DEFAULT_NIST_NVD_CVE_HISTORY_URL,
            token=token,
            timeout=timeout,
            rate_limit=rate_limit,
            request_attempts=request_attempts,
        )

    def changes(
        self,
        *,
        change_start_date: Optional[datetime] = None,
        change_end_date: Optional[datetime] = None,
        cve_id: Optional[str] = None,
        event_name: Optional[Union[EventName, str]] = None,
        request_results: Optional[int] = None,
        start_index: int = 0,
        results_per_page: Optional[int] = None,
    ) -> NVDResults[CVEChange]:
        """
        Get all CVEs for the provided arguments

        https://nvd.nist.gov/developers/vulnerabilities#divGetCves

        Args:
            change_start_date: Return all CVE changes after this date.
            change_end_date: Return all CVE changes before this date.
            cve_id: Return all CVE changes for this Common Vulnerabilities and Exposures identifier.
            event_name: Return all CVE changes with this event name.
            request_results: Number of CVEs changes to download. Set to None
                (default) to download all available CPEs.
            start_index: Index of the first CVE change to be returned. Useful
                only for paginated requests that should not start at the first
                page.
            results_per_page: Number of results in a single requests. Mostly
                useful for paginated requests.

        Returns:
            A NVDResponse for CVE changes

        Example:
            .. code-block:: python

                from pontos.nvd.cve_changes import CVEChangesApi

                async with CVEChangesApi() as api:
                    async for cve_change in api.changes(event_name=EventName.INITIAL_ANALYSIS):
                        print(cve_change)

                    json = api.changes(event_name=EventName.INITIAL_ANALYSIS).json()

                    async for changes in api.changes(
                        event_name=EventName.INITIAL_ANALYSIS,
                    ).chunks():
                        for cve_change in changes:
                        print(cve_change)
        """
        if change_start_date and not change_end_date:
            change_end_date = min(
                now(), change_start_date + timedelta(days=120)
            )
        elif change_end_date and not change_start_date:
            change_start_date = change_end_date - timedelta(days=120)

        params: Params = {}
        if change_start_date and change_end_date:
            if change_end_date - change_start_date > timedelta(days=120):
                raise PontosError(
                    "change_start_date and change_end_date must not be more than 120 days apart"
                )

            params["changeStartDate"] = format_date(change_start_date)
            params["changeEndDate"] = format_date(change_end_date)

        if cve_id:
            params["cveId"] = cve_id

        if event_name:
            params["eventName"] = event_name

        results_per_page = min(
            results_per_page or MAX_CVE_CHANGES_PER_PAGE,
            request_results or MAX_CVE_CHANGES_PER_PAGE,
        )
        return NVDResults(
            self,
            params,
            _result_iterator,
            request_results=request_results,
            results_per_page=results_per_page,
            start_index=start_index,
        )

    async def __aenter__(self) -> "CVEChangesApi":
        await super().__aenter__()
        return self



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

""" Argument parser for pontos-github-actions """

from argparse import ArgumentParser, Namespace
from typing import Optional, Sequence

import shtab

from .cmds import actions_input, actions_output


def split_pairs(value: str):
    if "=" not in value:
        raise ValueError(f"Must contain a 'name=value' pair not '{value}'.")
    return tuple(value.split("=", 1))


def parse_args(args: Optional[Sequence[str]] = None) -> Namespace:
    """
    Parsing args for Pontos GitHub Actions
    """

    parser = ArgumentParser(
        description="Greenbone GitHub Actions API.",
    )
    shtab.add_argument_to(parser)

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        required=True,
        help="additional help",
        dest="command",
    )

    output_parser = subparsers.add_parser("output", help="Set output variables")
    output_parser.add_argument(
        "output", help="Output as name=value pairs", type=split_pairs, nargs="+"
    )

    output_parser.set_defaults(func=actions_output)

    input_parser = subparsers.add_parser("input", help="Print input variables")
    input_parser.add_argument(
        "input",
        help="Name of the input variable to print",
        nargs="+",
    )
    input_parser.add_argument("--format", choices=["json"])

    input_parser.set_defaults(func=actions_input)

    return parser.parse_args(args)



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long
# ruff: noqa: E501

import unittest
from datetime import datetime, timezone

from pontos.github.models.tag import GitObjectType, Tag, VerificationReason


class TagModelTestCase(unittest.TestCase):
    def test_from_dict(self):
        data = {
            "node_id": "MDM6VGFnOTQwYmQzMzYyNDhlZmFlMGY5ZWU1YmM3YjJkNWM5ODU4ODdiMTZhYw==",
            "tag": "v0.0.1",
            "sha": "940bd336248efae0f9ee5bc7b2d5c985887b16ac",
            "url": "https://api.github.com/repos/octocat/Hello-World/git/tags/940bd336248efae0f9ee5bc7b2d5c985887b16ac",
            "message": "initial version",
            "tagger": {
                "name": "Monalisa Octocat",
                "email": "octocat@github.com",
                "date": "2014-11-07T22:01:45Z",
            },
            "object": {
                "type": "commit",
                "sha": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
                "url": "https://api.github.com/repos/octocat/Hello-World/git/commits/c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
            },
            "verification": {
                "verified": False,
                "reason": "unsigned",
                "signature": None,
                "payload": None,
            },
        }

        tag = Tag.from_dict(data)

        self.assertEqual(
            tag.node_id,
            "MDM6VGFnOTQwYmQzMzYyNDhlZmFlMGY5ZWU1YmM3YjJkNWM5ODU4ODdiMTZhYw==",
        )
        self.assertEqual(tag.tag, "v0.0.1")
        self.assertEqual(tag.sha, "940bd336248efae0f9ee5bc7b2d5c985887b16ac")
        self.assertEqual(
            tag.url,
            "https://api.github.com/repos/octocat/Hello-World/git/tags/940bd336248efae0f9ee5bc7b2d5c985887b16ac",
        )
        self.assertEqual(tag.message, "initial version")

        tagger = tag.tagger
        self.assertEqual(tagger.name, "Monalisa Octocat")
        self.assertEqual(tagger.email, "octocat@github.com")
        self.assertEqual(
            tagger.date, datetime(2014, 11, 7, 22, 1, 45, tzinfo=timezone.utc)
        )

        tag_object = tag.object
        self.assertEqual(tag_object.type, GitObjectType.COMMIT)
        self.assertEqual(
            tag_object.sha, "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c"
        )
        self.assertEqual(
            tag_object.url,
            "https://api.github.com/repos/octocat/Hello-World/git/commits/c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
        )

        verification = tag.verification
        self.assertFalse(verification.verified)
        self.assertEqual(verification.reason, VerificationReason.UNSIGNED)
        self.assertIsNone(verification.payload)
        self.assertIsNone(verification.signature)



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long, too-many-lines

import unittest
from datetime import datetime, timezone

from pontos.github.models.base import Permission, TeamPrivacy
from pontos.github.models.pull_request import (
    AuthorAssociation,
    Comment,
    MilestoneState,
    PullRequest,
    PullRequestState,
)


class PullRequestTestCase(unittest.TestCase):
    def test_from_dict(self):
        pr = PullRequest.from_dict(
            {
                "url": "https://api.github.com/repos/octocat/Hello-World/pulls/1347",
                "id": 1,
                "node_id": "MDExOlB1bGxSZXF1ZXN0MQ==",
                "html_url": "https://github.com/octocat/Hello-World/pull/1347",
                "diff_url": "https://github.com/octocat/Hello-World/pull/1347.diff",
                "patch_url": "https://github.com/octocat/Hello-World/pull/1347.patch",
                "issue_url": "https://api.github.com/repos/octocat/Hello-World/issues/1347",
                "commits_url": "https://api.github.com/repos/octocat/Hello-World/pulls/1347/commits",
                "review_comments_url": "https://api.github.com/repos/octocat/Hello-World/pulls/1347/comments",
                "review_comment_url": "https://api.github.com/repos/octocat/Hello-World/pulls/comments{/number}",
                "comments_url": "https://api.github.com/repos/octocat/Hello-World/issues/1347/comments",
                "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/6dcb09b5b57875f334f61aebed695e2e4193db5e",
                "number": 1347,
                "state": "open",
                "locked": True,
                "title": "Amazing new feature",
                "user": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "body": "Please pull these awesome changes in!",
                "labels": [
                    {
                        "id": 208045946,
                        "node_id": "MDU6TGFiZWwyMDgwNDU5NDY=",
                        "url": "https://api.github.com/repos/octocat/Hello-World/labels/bug",
                        "name": "bug",
                        "description": "Something isn't working",
                        "color": "f29513",
                        "default": True,
                    }
                ],
                "milestone": {
                    "url": "https://api.github.com/repos/octocat/Hello-World/milestones/1",
                    "html_url": "https://github.com/octocat/Hello-World/milestones/v1.0",
                    "labels_url": "https://api.github.com/repos/octocat/Hello-World/milestones/1/labels",
                    "id": 1002604,
                    "node_id": "MDk6TWlsZXN0b25lMTAwMjYwNA==",
                    "number": 1,
                    "state": "open",
                    "title": "v1.0",
                    "description": "Tracking milestone for version 1.0",
                    "creator": {
                        "login": "octocat",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/octocat",
                        "html_url": "https://github.com/octocat",
                        "followers_url": "https://api.github.com/users/octocat/followers",
                        "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                        "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                        "organizations_url": "https://api.github.com/users/octocat/orgs",
                        "repos_url": "https://api.github.com/users/octocat/repos",
                        "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/octocat/received_events",
                        "type": "User",
                        "site_admin": False,
                    },
                    "open_issues": 4,
                    "closed_issues": 8,
                    "created_at": "2011-04-10T20:09:31Z",
                    "updated_at": "2014-03-03T18:58:10Z",
                    "closed_at": "2013-02-12T13:22:01Z",
                    "due_on": "2012-10-09T23:39:01Z",
                },
                "active_lock_reason": "too heated",
                "created_at": "2011-01-26T19:01:12Z",
                "updated_at": "2011-01-26T19:01:12Z",
                "closed_at": "2011-01-26T19:01:12Z",
                "merged_at": "2011-01-26T19:01:12Z",
                "merge_commit_sha": "e5bd3914e2e596debea16f433f57875b5b90bcd6",
                "assignee": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "assignees": [
                    {
                        "login": "octocat",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/octocat",
                        "html_url": "https://github.com/octocat",
                        "followers_url": "https://api.github.com/users/octocat/followers",
                        "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                        "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                        "organizations_url": "https://api.github.com/users/octocat/orgs",
                        "repos_url": "https://api.github.com/users/octocat/repos",
                        "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/octocat/received_events",
                        "type": "User",
                        "site_admin": False,
                    },
                    {
                        "login": "hubot",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/hubot_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/hubot",
                        "html_url": "https://github.com/hubot",
                        "followers_url": "https://api.github.com/users/hubot/followers",
                        "following_url": "https://api.github.com/users/hubot/following{/other_user}",
                        "gists_url": "https://api.github.com/users/hubot/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/hubot/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/hubot/subscriptions",
                        "organizations_url": "https://api.github.com/users/hubot/orgs",
                        "repos_url": "https://api.github.com/users/hubot/repos",
                        "events_url": "https://api.github.com/users/hubot/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/hubot/received_events",
                        "type": "User",
                        "site_admin": True,
                    },
                ],
                "requested_reviewers": [
                    {
                        "login": "other_user",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/other_user_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/other_user",
                        "html_url": "https://github.com/other_user",
                        "followers_url": "https://api.github.com/users/other_user/followers",
                        "following_url": "https://api.github.com/users/other_user/following{/other_user}",
                        "gists_url": "https://api.github.com/users/other_user/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/other_user/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/other_user/subscriptions",
                        "organizations_url": "https://api.github.com/users/other_user/orgs",
                        "repos_url": "https://api.github.com/users/other_user/repos",
                        "events_url": "https://api.github.com/users/other_user/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/other_user/received_events",
                        "type": "User",
                        "site_admin": False,
                    }
                ],
                "requested_teams": [
                    {
                        "id": 1,
                        "node_id": "MDQ6VGVhbTE=",
                        "url": "https://api.github.com/teams/1",
                        "html_url": "https://github.com/orgs/github/teams/justice-league",
                        "name": "Justice League",
                        "slug": "justice-league",
                        "description": "A great team.",
                        "privacy": "closed",
                        "permission": "admin",
                        "members_url": "https://api.github.com/teams/1/members{/member}",
                        "repositories_url": "https://api.github.com/teams/1/repos",
                    }
                ],
                "head": {
                    "label": "octocat:new-topic",
                    "ref": "new-topic",
                    "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
                    "user": {
                        "login": "octocat",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/octocat",
                        "html_url": "https://github.com/octocat",
                        "followers_url": "https://api.github.com/users/octocat/followers",
                        "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                        "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                        "organizations_url": "https://api.github.com/users/octocat/orgs",
                        "repos_url": "https://api.github.com/users/octocat/repos",
                        "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/octocat/received_events",
                        "type": "User",
                        "site_admin": False,
                    },
                    "repo": {
                        "id": 1296269,
                        "node_id": "MDEwOlJlcG9zaXRvcnkxMjk2MjY5",
                        "name": "Hello-World",
                        "full_name": "octocat/Hello-World",
                        "owner": {
                            "login": "octocat",
                            "id": 1,
                            "node_id": "MDQ6VXNlcjE=",
                            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                            "gravatar_id": "",
                            "url": "https://api.github.com/users/octocat",
                            "html_url": "https://github.com/octocat",
                            "followers_url": "https://api.github.com/users/octocat/followers",
                            "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                            "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                            "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                            "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                            "organizations_url": "https://api.github.com/users/octocat/orgs",
                            "repos_url": "https://api.github.com/users/octocat/repos",
                            "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                            "received_events_url": "https://api.github.com/users/octocat/received_events",
                            "type": "User",
                            "site_admin": False,
                        },
                        "private": False,
                        "html_url": "https://github.com/octocat/Hello-World",
                        "description": "This your first repo!",
                        "fork": False,
                        "url": "https://api.github.com/repos/octocat/Hello-World",
                        "archive_url": "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
                        "assignees_url": "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
                        "blobs_url": "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
                        "branches_url": "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
                        "collaborators_url": "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
                        "comments_url": "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
                        "commits_url": "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
                        "compare_url": "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
                        "contents_url": "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
                        "contributors_url": "https://api.github.com/repos/octocat/Hello-World/contributors",
                        "deployments_url": "https://api.github.com/repos/octocat/Hello-World/deployments",
                        "downloads_url": "https://api.github.com/repos/octocat/Hello-World/downloads",
                        "events_url": "https://api.github.com/repos/octocat/Hello-World/events",
                        "forks_url": "https://api.github.com/repos/octocat/Hello-World/forks",
                        "git_commits_url": "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
                        "git_refs_url": "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
                        "git_tags_url": "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
                        "git_url": "git:github.com/octocat/Hello-World.git",
                        "issue_comment_url": "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
                        "issue_events_url": "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
                        "issues_url": "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
                        "keys_url": "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
                        "labels_url": "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
                        "languages_url": "https://api.github.com/repos/octocat/Hello-World/languages",
                        "merges_url": "https://api.github.com/repos/octocat/Hello-World/merges",
                        "milestones_url": "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
                        "notifications_url": "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
                        "pulls_url": "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
                        "releases_url": "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
                        "ssh_url": "git@github.com:octocat/Hello-World.git",
                        "stargazers_url": "https://api.github.com/repos/octocat/Hello-World/stargazers",
                        "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
                        "subscribers_url": "https://api.github.com/repos/octocat/Hello-World/subscribers",
                        "subscription_url": "https://api.github.com/repos/octocat/Hello-World/subscription",
                        "tags_url": "https://api.github.com/repos/octocat/Hello-World/tags",
                        "teams_url": "https://api.github.com/repos/octocat/Hello-World/teams",
                        "trees_url": "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
                        "clone_url": "https://github.com/octocat/Hello-World.git",
                        "mirror_url": "git:git.example.com/octocat/Hello-World",
                        "hooks_url": "https://api.github.com/repos/octocat/Hello-World/hooks",
                        "svn_url": "https://svn.github.com/octocat/Hello-World",
                        "homepage": "https://github.com",
                        "language": None,
                        "forks_count": 9,
                        "stargazers_count": 80,
                        "watchers_count": 80,
                        "size": 108,
                        "default_branch": "master",
                        "open_issues_count": 0,
                        "topics": ["octocat", "atom", "electron", "api"],
                        "has_issues": True,
                        "has_projects": True,
                        "has_wiki": True,
                        "has_pages": False,
                        "has_downloads": True,
                        "has_discussions": False,
                        "archived": False,
                        "disabled": False,
                        "pushed_at": "2011-01-26T19:06:43Z",
                        "created_at": "2011-01-26T19:01:12Z",
                        "updated_at": "2011-01-26T19:14:43Z",
                        "permissions": {
                            "admin": False,
                            "push": False,
                            "pull": True,
                        },
                        "allow_rebase_merge": True,
                        "allow_squash_merge": True,
                        "allow_merge_commit": True,
                        "allow_forking": True,
                        "forks": 123,
                        "open_issues": 123,
                        "license": {
                            "key": "mit",
                            "name": "MIT License",
                            "url": "https://api.github.com/licenses/mit",
                            "spdx_id": "MIT",
                            "node_id": "MDc6TGljZW5zZW1pdA==",
                        },
                        "watchers": 123,
                    },
                },
                "base": {
                    "label": "octocat:master",
                    "ref": "master",
                    "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
                    "user": {
                        "login": "octocat",
                        "id": 1,
                        "node_id": "MDQ6VXNlcjE=",
                        "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/octocat",
                        "html_url": "https://github.com/octocat",
                        "followers_url": "https://api.github.com/users/octocat/followers",
                        "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                        "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                        "organizations_url": "https://api.github.com/users/octocat/orgs",
                        "repos_url": "https://api.github.com/users/octocat/repos",
                        "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/octocat/received_events",
                        "type": "User",
                        "site_admin": False,
                    },
                    "repo": {
                        "id": 1296269,
                        "node_id": "MDEwOlJlcG9zaXRvcnkxMjk2MjY5",
                        "name": "Hello-World",
                        "full_name": "octocat/Hello-World",
                        "owner": {
                            "login": "octocat",
                            "id": 1,
                            "node_id": "MDQ6VXNlcjE=",
                            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                            "gravatar_id": "",
                            "url": "https://api.github.com/users/octocat",
                            "html_url": "https://github.com/octocat",
                            "followers_url": "https://api.github.com/users/octocat/followers",
                            "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                            "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                            "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                            "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                            "organizations_url": "https://api.github.com/users/octocat/orgs",
                            "repos_url": "https://api.github.com/users/octocat/repos",
                            "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                            "received_events_url": "https://api.github.com/users/octocat/received_events",
                            "type": "User",
                            "site_admin": False,
                        },
                        "private": False,
                        "html_url": "https://github.com/octocat/Hello-World",
                        "description": "This your first repo!",
                        "fork": False,
                        "url": "https://api.github.com/repos/octocat/Hello-World",
                        "archive_url": "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
                        "assignees_url": "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
                        "blobs_url": "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
                        "branches_url": "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
                        "collaborators_url": "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
                        "comments_url": "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
                        "commits_url": "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
                        "compare_url": "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
                        "contents_url": "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
                        "contributors_url": "https://api.github.com/repos/octocat/Hello-World/contributors",
                        "deployments_url": "https://api.github.com/repos/octocat/Hello-World/deployments",
                        "downloads_url": "https://api.github.com/repos/octocat/Hello-World/downloads",
                        "events_url": "https://api.github.com/repos/octocat/Hello-World/events",
                        "forks_url": "https://api.github.com/repos/octocat/Hello-World/forks",
                        "git_commits_url": "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
                        "git_refs_url": "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
                        "git_tags_url": "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
                        "git_url": "git:github.com/octocat/Hello-World.git",
                        "issue_comment_url": "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
                        "issue_events_url": "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
                        "issues_url": "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
                        "keys_url": "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
                        "labels_url": "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
                        "languages_url": "https://api.github.com/repos/octocat/Hello-World/languages",
                        "merges_url": "https://api.github.com/repos/octocat/Hello-World/merges",
                        "milestones_url": "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
                        "notifications_url": "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
                        "pulls_url": "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
                        "releases_url": "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
                        "ssh_url": "git@github.com:octocat/Hello-World.git",
                        "stargazers_url": "https://api.github.com/repos/octocat/Hello-World/stargazers",
                        "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
                        "subscribers_url": "https://api.github.com/repos/octocat/Hello-World/subscribers",
                        "subscription_url": "https://api.github.com/repos/octocat/Hello-World/subscription",
                        "tags_url": "https://api.github.com/repos/octocat/Hello-World/tags",
                        "teams_url": "https://api.github.com/repos/octocat/Hello-World/teams",
                        "trees_url": "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
                        "clone_url": "https://github.com/octocat/Hello-World.git",
                        "mirror_url": "git:git.example.com/octocat/Hello-World",
                        "hooks_url": "https://api.github.com/repos/octocat/Hello-World/hooks",
                        "svn_url": "https://svn.github.com/octocat/Hello-World",
                        "homepage": "https://github.com",
                        "language": None,
                        "forks_count": 9,
                        "stargazers_count": 80,
                        "watchers_count": 80,
                        "size": 108,
                        "default_branch": "master",
                        "open_issues_count": 0,
                        "topics": ["octocat", "atom", "electron", "api"],
                        "has_issues": True,
                        "has_projects": True,
                        "has_wiki": True,
                        "has_pages": False,
                        "has_downloads": True,
                        "has_discussions": False,
                        "archived": False,
                        "disabled": False,
                        "pushed_at": "2011-01-26T19:06:43Z",
                        "created_at": "2011-01-26T19:01:12Z",
                        "updated_at": "2011-01-26T19:14:43Z",
                        "permissions": {
                            "admin": False,
                            "push": False,
                            "pull": True,
                        },
                        "allow_rebase_merge": True,
                        "temp_clone_token": "ABTLWHOULUVAXGTRYU7OC2876QJ2O",
                        "allow_squash_merge": True,
                        "allow_merge_commit": True,
                        "allow_forking": True,
                        "forks": 123,
                        "open_issues": 123,
                        "license": {
                            "key": "mit",
                            "name": "MIT License",
                            "url": "https://api.github.com/licenses/mit",
                            "spdx_id": "MIT",
                            "node_id": "MDc6TGljZW5zZW1pdA==",
                        },
                        "watchers": 123,
                    },
                },
                "_links": {
                    "self": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/pulls/1347"
                    },
                    "html": {
                        "href": "https://github.com/octocat/Hello-World/pull/1347"
                    },
                    "issue": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/issues/1347"
                    },
                    "comments": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/issues/1347/comments"
                    },
                    "review_comments": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/pulls/1347/comments"
                    },
                    "review_comment": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/pulls/comments{/number}"
                    },
                    "commits": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/pulls/1347/commits"
                    },
                    "statuses": {
                        "href": "https://api.github.com/repos/octocat/Hello-World/statuses/6dcb09b5b57875f334f61aebed695e2e4193db5e"
                    },
                },
                "author_association": "OWNER",
                "auto_merge": None,
                "draft": False,
                "merged": False,
                "mergeable": True,
                "rebaseable": True,
                "mergeable_state": "clean",
                "merged_by": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "comments": 10,
                "review_comments": 0,
                "maintainer_can_modify": True,
                "commits": 3,
                "additions": 100,
                "deletions": 3,
                "changed_files": 5,
            }
        )

        self.assertEqual(
            pr.url,
            "https://api.github.com/repos/octocat/Hello-World/pulls/1347",
        )
        self.assertEqual(pr.id, 1)
        self.assertEqual(pr.node_id, "MDExOlB1bGxSZXF1ZXN0MQ==")
        self.assertEqual(
            pr.html_url, "https://github.com/octocat/Hello-World/pull/1347"
        )
        self.assertEqual(
            pr.diff_url, "https://github.com/octocat/Hello-World/pull/1347.diff"
        )
        self.assertEqual(
            pr.patch_url,
            "https://github.com/octocat/Hello-World/pull/1347.patch",
        )
        self.assertEqual(
            pr.issue_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/1347",
        )
        self.assertEqual(
            pr.commits_url,
            "https://api.github.com/repos/octocat/Hello-World/pulls/1347/commits",
        )
        self.assertEqual(
            pr.review_comments_url,
            "https://api.github.com/repos/octocat/Hello-World/pulls/1347/comments",
        )
        self.assertEqual(
            pr.review_comment_url,
            "https://api.github.com/repos/octocat/Hello-World/pulls/comments{/number}",
        )
        self.assertEqual(
            pr.comments_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/1347/comments",
        )
        self.assertEqual(
            pr.statuses_url,
            "https://api.github.com/repos/octocat/Hello-World/statuses/6dcb09b5b57875f334f61aebed695e2e4193db5e",
        )
        self.assertEqual(pr.number, 1347)
        self.assertEqual(pr.state, PullRequestState.OPEN)
        self.assertTrue(pr.locked)
        self.assertEqual(pr.title, "Amazing new feature")
        self.assertEqual(pr.body, "Please pull these awesome changes in!")
        self.assertEqual(pr.active_lock_reason, "too heated")
        self.assertEqual(
            pr.created_at, datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc)
        )
        self.assertEqual(
            pr.updated_at, datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc)
        )
        self.assertEqual(
            pr.closed_at, datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc)
        )
        self.assertEqual(
            pr.merged_at, datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc)
        )
        self.assertEqual(
            pr.merge_commit_sha, "e5bd3914e2e596debea16f433f57875b5b90bcd6"
        )
        self.assertEqual(pr.author_association, AuthorAssociation.OWNER)
        self.assertEqual(pr.auto_merge, None)
        self.assertEqual(pr.draft, False)
        self.assertEqual(pr.merged, False)
        self.assertEqual(pr.mergeable, True)
        self.assertEqual(pr.rebaseable, True)
        self.assertEqual(pr.mergeable_state, "clean")
        self.assertEqual(pr.comments, 10)
        self.assertEqual(pr.review_comments, 0)
        self.assertTrue(pr.maintainer_can_modify)
        self.assertEqual(pr.commits, 3)
        self.assertEqual(pr.additions, 100)
        self.assertEqual(pr.deletions, 3)
        self.assertEqual(pr.changed_files, 5)

        user = pr.user
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        self.assertEqual(len(pr.labels), 1)
        label = pr.labels[0]
        self.assertEqual(label.id, 208045946)
        self.assertEqual(label.node_id, "MDU6TGFiZWwyMDgwNDU5NDY=")
        self.assertEqual(
            label.url,
            "https://api.github.com/repos/octocat/Hello-World/labels/bug",
        )
        self.assertEqual(label.name, "bug")
        self.assertEqual(label.description, "Something isn't working")
        self.assertEqual(label.color, "f29513")
        self.assertTrue(label.default)

        milestone = pr.milestone
        self.assertEqual(
            milestone.url,
            "https://api.github.com/repos/octocat/Hello-World/milestones/1",
        )
        self.assertEqual(
            milestone.html_url,
            "https://github.com/octocat/Hello-World/milestones/v1.0",
        )
        self.assertEqual(
            milestone.labels_url,
            "https://api.github.com/repos/octocat/Hello-World/milestones/1/labels",
        )
        self.assertEqual(milestone.id, 1002604)
        self.assertEqual(milestone.node_id, "MDk6TWlsZXN0b25lMTAwMjYwNA==")
        self.assertEqual(milestone.number, 1)
        self.assertEqual(milestone.state, MilestoneState.OPEN)
        self.assertEqual(milestone.title, "v1.0")
        self.assertEqual(
            milestone.description, "Tracking milestone for version 1.0"
        )
        self.assertEqual(milestone.open_issues, 4)
        self.assertEqual(milestone.closed_issues, 8)
        self.assertEqual(
            milestone.created_at,
            datetime(2011, 4, 10, 20, 9, 31, tzinfo=timezone.utc),
        )
        self.assertEqual(
            milestone.updated_at,
            datetime(2014, 3, 3, 18, 58, 10, tzinfo=timezone.utc),
        )
        self.assertEqual(
            milestone.closed_at,
            datetime(2013, 2, 12, 13, 22, 1, tzinfo=timezone.utc),
        )
        self.assertEqual(
            milestone.due_on,
            datetime(2012, 10, 9, 23, 39, 1, tzinfo=timezone.utc),
        )

        creator = milestone.creator
        self.assertEqual(creator.login, "octocat")
        self.assertEqual(creator.id, 1)
        self.assertEqual(creator.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            creator.avatar_url,
            "https://github.com/images/error/octocat_happy.gif",
        )
        self.assertEqual(creator.gravatar_id, "")
        self.assertEqual(creator.url, "https://api.github.com/users/octocat")
        self.assertEqual(creator.html_url, "https://github.com/octocat")
        self.assertEqual(
            creator.followers_url,
            "https://api.github.com/users/octocat/followers",
        )
        self.assertEqual(
            creator.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            creator.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            creator.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            creator.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            creator.organizations_url,
            "https://api.github.com/users/octocat/orgs",
        )
        self.assertEqual(
            creator.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            creator.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            creator.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(creator.type, "User")
        self.assertEqual(creator.site_admin, False)

        user = pr.assignee
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        self.assertEqual(len(pr.assignees), 2)
        user = pr.assignees[0]
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        user = pr.assignees[1]
        self.assertEqual(user.login, "hubot")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/hubot_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/hubot")
        self.assertEqual(user.html_url, "https://github.com/hubot")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/hubot/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/hubot/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url, "https://api.github.com/users/hubot/gists{/gist_id}"
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/hubot/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/hubot/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/hubot/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/hubot/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/hubot/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/hubot/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertTrue(user.site_admin)

        self.assertEqual(len(pr.requested_reviewers), 1)
        reviewer = pr.requested_reviewers[0]
        self.assertEqual(reviewer.login, "other_user")
        self.assertEqual(reviewer.id, 1)
        self.assertEqual(reviewer.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            reviewer.avatar_url,
            "https://github.com/images/error/other_user_happy.gif",
        )
        self.assertEqual(reviewer.gravatar_id, "")
        self.assertEqual(
            reviewer.url, "https://api.github.com/users/other_user"
        )
        self.assertEqual(reviewer.html_url, "https://github.com/other_user")
        self.assertEqual(
            reviewer.followers_url,
            "https://api.github.com/users/other_user/followers",
        )
        self.assertEqual(
            reviewer.following_url,
            "https://api.github.com/users/other_user/following{/other_user}",
        )
        self.assertEqual(
            reviewer.gists_url,
            "https://api.github.com/users/other_user/gists{/gist_id}",
        )
        self.assertEqual(
            reviewer.starred_url,
            "https://api.github.com/users/other_user/starred{/owner}{/repo}",
        )
        self.assertEqual(
            reviewer.subscriptions_url,
            "https://api.github.com/users/other_user/subscriptions",
        )
        self.assertEqual(
            reviewer.organizations_url,
            "https://api.github.com/users/other_user/orgs",
        )
        self.assertEqual(
            reviewer.repos_url, "https://api.github.com/users/other_user/repos"
        )
        self.assertEqual(
            reviewer.events_url,
            "https://api.github.com/users/other_user/events{/privacy}",
        )
        self.assertEqual(
            reviewer.received_events_url,
            "https://api.github.com/users/other_user/received_events",
        )
        self.assertEqual(reviewer.type, "User")
        self.assertEqual(reviewer.site_admin, False)

        self.assertEqual(len(pr.requested_teams), 1)
        team = pr.requested_teams[0]
        self.assertEqual(team.id, 1)
        self.assertEqual(team.node_id, "MDQ6VGVhbTE=")
        self.assertEqual(team.url, "https://api.github.com/teams/1")
        self.assertEqual(
            team.html_url, "https://github.com/orgs/github/teams/justice-league"
        )
        self.assertEqual(team.name, "Justice League")
        self.assertEqual(team.slug, "justice-league")
        self.assertEqual(team.description, "A great team.")
        self.assertEqual(team.privacy, TeamPrivacy.CLOSED)
        self.assertEqual(team.permission, Permission.ADMIN)
        self.assertEqual(
            team.members_url, "https://api.github.com/teams/1/members{/member}"
        )
        self.assertEqual(
            team.repositories_url, "https://api.github.com/teams/1/repos"
        )

        head = pr.head
        self.assertEqual(head.label, "octocat:new-topic")
        self.assertEqual(head.ref, "new-topic")
        self.assertEqual(head.sha, "6dcb09b5b57875f334f61aebed695e2e4193db5e")

        user = head.user
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        repo = head.repo
        self.assertEqual(repo.id, 1296269)
        self.assertEqual(repo.node_id, "MDEwOlJlcG9zaXRvcnkxMjk2MjY5")
        self.assertEqual(repo.name, "Hello-World")
        self.assertEqual(repo.full_name, "octocat/Hello-World")
        self.assertFalse(repo.private)
        self.assertEqual(
            repo.html_url, "https://github.com/octocat/Hello-World"
        )
        self.assertEqual(repo.description, "This your first repo!")
        self.assertEqual(repo.fork, False)
        self.assertEqual(
            repo.url, "https://api.github.com/repos/octocat/Hello-World"
        )
        self.assertEqual(
            repo.archive_url,
            "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
        )
        self.assertEqual(
            repo.assignees_url,
            "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
        )
        self.assertEqual(
            repo.blobs_url,
            "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
        )
        self.assertEqual(
            repo.branches_url,
            "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
        )
        self.assertEqual(
            repo.collaborators_url,
            "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
        )
        self.assertEqual(
            repo.comments_url,
            "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
        )
        self.assertEqual(
            repo.commits_url,
            "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
        )
        self.assertEqual(
            repo.compare_url,
            "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
        )
        self.assertEqual(
            repo.contents_url,
            "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
        )
        self.assertEqual(
            repo.contributors_url,
            "https://api.github.com/repos/octocat/Hello-World/contributors",
        )
        self.assertEqual(
            repo.deployments_url,
            "https://api.github.com/repos/octocat/Hello-World/deployments",
        )
        self.assertEqual(
            repo.downloads_url,
            "https://api.github.com/repos/octocat/Hello-World/downloads",
        )
        self.assertEqual(
            repo.events_url,
            "https://api.github.com/repos/octocat/Hello-World/events",
        )
        self.assertEqual(
            repo.forks_url,
            "https://api.github.com/repos/octocat/Hello-World/forks",
        )
        self.assertEqual(
            repo.git_commits_url,
            "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
        )
        self.assertEqual(
            repo.git_refs_url,
            "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
        )
        self.assertEqual(
            repo.git_tags_url,
            "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
        )
        self.assertEqual(repo.git_url, "git:github.com/octocat/Hello-World.git")
        self.assertEqual(
            repo.issue_comment_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
        )
        self.assertEqual(
            repo.issue_events_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
        )
        self.assertEqual(
            repo.issues_url,
            "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
        )
        self.assertEqual(
            repo.keys_url,
            "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
        )
        self.assertEqual(
            repo.labels_url,
            "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
        )
        self.assertEqual(
            repo.languages_url,
            "https://api.github.com/repos/octocat/Hello-World/languages",
        )
        self.assertEqual(
            repo.merges_url,
            "https://api.github.com/repos/octocat/Hello-World/merges",
        )
        self.assertEqual(
            repo.milestones_url,
            "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
        )
        self.assertEqual(
            repo.notifications_url,
            "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
        )
        self.assertEqual(
            repo.pulls_url,
            "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
        )
        self.assertEqual(
            repo.releases_url,
            "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
        )
        self.assertEqual(repo.ssh_url, "git@github.com:octocat/Hello-World.git")
        self.assertEqual(
            repo.stargazers_url,
            "https://api.github.com/repos/octocat/Hello-World/stargazers",
        )
        self.assertEqual(
            repo.statuses_url,
            "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
        )
        self.assertEqual(
            repo.subscribers_url,
            "https://api.github.com/repos/octocat/Hello-World/subscribers",
        )
        self.assertEqual(
            repo.subscription_url,
            "https://api.github.com/repos/octocat/Hello-World/subscription",
        )
        self.assertEqual(
            repo.tags_url,
            "https://api.github.com/repos/octocat/Hello-World/tags",
        )
        self.assertEqual(
            repo.teams_url,
            "https://api.github.com/repos/octocat/Hello-World/teams",
        )
        self.assertEqual(
            repo.trees_url,
            "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
        )
        self.assertEqual(
            repo.clone_url, "https://github.com/octocat/Hello-World.git"
        )
        self.assertEqual(
            repo.mirror_url, "git:git.example.com/octocat/Hello-World"
        )
        self.assertEqual(
            repo.hooks_url,
            "https://api.github.com/repos/octocat/Hello-World/hooks",
        )
        self.assertEqual(
            repo.svn_url, "https://svn.github.com/octocat/Hello-World"
        )
        self.assertEqual(repo.homepage, "https://github.com")
        self.assertIsNone(repo.language)
        self.assertEqual(repo.forks_count, 9)
        self.assertEqual(repo.stargazers_count, 80)
        self.assertEqual(repo.watchers_count, 80)
        self.assertEqual(repo.size, 108)
        self.assertEqual(repo.default_branch, "master")
        self.assertEqual(repo.open_issues_count, 0)
        self.assertEqual(repo.topics, ["octocat", "atom", "electron", "api"])
        self.assertTrue(repo.has_issues)
        self.assertTrue(repo.has_projects)
        self.assertTrue(repo.has_wiki)
        self.assertFalse(repo.has_pages)
        self.assertTrue(repo.has_downloads)
        self.assertFalse(repo.has_discussions)
        self.assertFalse(repo.archived)
        self.assertFalse(repo.disabled)
        self.assertEqual(
            repo.pushed_at,
            datetime(2011, 1, 26, 19, 6, 43, tzinfo=timezone.utc),
        )
        self.assertEqual(
            repo.created_at,
            datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc),
        )
        self.assertEqual(
            repo.updated_at,
            datetime(2011, 1, 26, 19, 14, 43, tzinfo=timezone.utc),
        )
        self.assertTrue(repo.allow_rebase_merge)
        self.assertTrue(repo.allow_squash_merge)
        self.assertTrue(repo.allow_merge_commit)
        self.assertTrue(repo.allow_forking)
        self.assertEqual(repo.forks, 123)
        self.assertEqual(repo.open_issues, 123)
        self.assertEqual(repo.watchers, 123)

        user = repo.owner
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        license = repo.license  # pylint: disable=redefined-builtin
        self.assertEqual(license.key, "mit")
        self.assertEqual(license.name, "MIT License")
        self.assertEqual(license.url, "https://api.github.com/licenses/mit")
        self.assertEqual(license.spdx_id, "MIT")
        self.assertEqual(license.node_id, "MDc6TGljZW5zZW1pdA==")

        self.assertFalse(repo.permissions.admin)
        self.assertFalse(repo.permissions.push)
        self.assertTrue(repo.permissions.pull)

        base = pr.base
        self.assertEqual(base.label, "octocat:master")
        self.assertEqual(base.ref, "master")
        self.assertEqual(base.sha, "6dcb09b5b57875f334f61aebed695e2e4193db5e")

        user = base.user
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        repo = base.repo
        self.assertEqual(repo.id, 1296269)
        self.assertEqual(repo.node_id, "MDEwOlJlcG9zaXRvcnkxMjk2MjY5")
        self.assertEqual(repo.name, "Hello-World")
        self.assertEqual(repo.full_name, "octocat/Hello-World")
        self.assertFalse(repo.private)
        self.assertEqual(
            repo.html_url, "https://github.com/octocat/Hello-World"
        )
        self.assertEqual(repo.description, "This your first repo!")
        self.assertEqual(repo.fork, False)
        self.assertEqual(
            repo.url, "https://api.github.com/repos/octocat/Hello-World"
        )
        self.assertEqual(
            repo.archive_url,
            "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
        )
        self.assertEqual(
            repo.assignees_url,
            "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
        )
        self.assertEqual(
            repo.blobs_url,
            "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
        )
        self.assertEqual(
            repo.branches_url,
            "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
        )
        self.assertEqual(
            repo.collaborators_url,
            "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
        )
        self.assertEqual(
            repo.comments_url,
            "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
        )
        self.assertEqual(
            repo.commits_url,
            "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
        )
        self.assertEqual(
            repo.compare_url,
            "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
        )
        self.assertEqual(
            repo.contents_url,
            "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
        )
        self.assertEqual(
            repo.contributors_url,
            "https://api.github.com/repos/octocat/Hello-World/contributors",
        )
        self.assertEqual(
            repo.deployments_url,
            "https://api.github.com/repos/octocat/Hello-World/deployments",
        )
        self.assertEqual(
            repo.downloads_url,
            "https://api.github.com/repos/octocat/Hello-World/downloads",
        )
        self.assertEqual(
            repo.events_url,
            "https://api.github.com/repos/octocat/Hello-World/events",
        )
        self.assertEqual(
            repo.forks_url,
            "https://api.github.com/repos/octocat/Hello-World/forks",
        )
        self.assertEqual(
            repo.git_commits_url,
            "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
        )
        self.assertEqual(
            repo.git_refs_url,
            "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
        )
        self.assertEqual(
            repo.git_tags_url,
            "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
        )
        self.assertEqual(repo.git_url, "git:github.com/octocat/Hello-World.git")
        self.assertEqual(
            repo.issue_comment_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
        )
        self.assertEqual(
            repo.issue_events_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
        )
        self.assertEqual(
            repo.issues_url,
            "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
        )
        self.assertEqual(
            repo.keys_url,
            "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
        )
        self.assertEqual(
            repo.labels_url,
            "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
        )
        self.assertEqual(
            repo.languages_url,
            "https://api.github.com/repos/octocat/Hello-World/languages",
        )
        self.assertEqual(
            repo.merges_url,
            "https://api.github.com/repos/octocat/Hello-World/merges",
        )
        self.assertEqual(
            repo.milestones_url,
            "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
        )
        self.assertEqual(
            repo.notifications_url,
            "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
        )
        self.assertEqual(
            repo.pulls_url,
            "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
        )
        self.assertEqual(
            repo.releases_url,
            "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
        )
        self.assertEqual(repo.ssh_url, "git@github.com:octocat/Hello-World.git")
        self.assertEqual(
            repo.stargazers_url,
            "https://api.github.com/repos/octocat/Hello-World/stargazers",
        )
        self.assertEqual(
            repo.statuses_url,
            "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
        )
        self.assertEqual(
            repo.subscribers_url,
            "https://api.github.com/repos/octocat/Hello-World/subscribers",
        )
        self.assertEqual(
            repo.subscription_url,
            "https://api.github.com/repos/octocat/Hello-World/subscription",
        )
        self.assertEqual(
            repo.tags_url,
            "https://api.github.com/repos/octocat/Hello-World/tags",
        )
        self.assertEqual(
            repo.teams_url,
            "https://api.github.com/repos/octocat/Hello-World/teams",
        )
        self.assertEqual(
            repo.trees_url,
            "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
        )
        self.assertEqual(
            repo.clone_url, "https://github.com/octocat/Hello-World.git"
        )
        self.assertEqual(
            repo.mirror_url, "git:git.example.com/octocat/Hello-World"
        )
        self.assertEqual(
            repo.hooks_url,
            "https://api.github.com/repos/octocat/Hello-World/hooks",
        )
        self.assertEqual(
            repo.svn_url, "https://svn.github.com/octocat/Hello-World"
        )
        self.assertEqual(repo.homepage, "https://github.com")
        self.assertIsNone(repo.language)
        self.assertEqual(repo.forks_count, 9)
        self.assertEqual(repo.stargazers_count, 80)
        self.assertEqual(repo.watchers_count, 80)
        self.assertEqual(repo.size, 108)
        self.assertEqual(repo.default_branch, "master")
        self.assertEqual(repo.open_issues_count, 0)
        self.assertEqual(repo.topics, ["octocat", "atom", "electron", "api"])
        self.assertTrue(repo.has_issues)
        self.assertTrue(repo.has_projects)
        self.assertTrue(repo.has_wiki)
        self.assertFalse(repo.has_pages)
        self.assertTrue(repo.has_downloads)
        self.assertFalse(repo.has_discussions)
        self.assertFalse(repo.archived)
        self.assertFalse(repo.disabled)
        self.assertEqual(
            repo.pushed_at,
            datetime(2011, 1, 26, 19, 6, 43, tzinfo=timezone.utc),
        )
        self.assertEqual(
            repo.created_at,
            datetime(2011, 1, 26, 19, 1, 12, tzinfo=timezone.utc),
        )
        self.assertEqual(
            repo.updated_at,
            datetime(2011, 1, 26, 19, 14, 43, tzinfo=timezone.utc),
        )
        self.assertTrue(repo.allow_rebase_merge)
        self.assertTrue(repo.allow_squash_merge)
        self.assertTrue(repo.allow_merge_commit)
        self.assertTrue(repo.allow_forking)
        self.assertEqual(repo.forks, 123)
        self.assertEqual(repo.open_issues, 123)
        self.assertEqual(repo.watchers, 123)

        user = pr.merged_by
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)


class CommentTestCase(unittest.TestCase):
    def test_from_dict(self):
        data = {
            "id": 1,
            "node_id": "MDEyOklzc3VlQ29tbWVudDE=",
            "url": "https://api.github.com/repos/octocat/Hello-World/issues/comments/1",
            "html_url": "https://github.com/octocat/Hello-World/issues/1347#issuecomment-1",
            "body": "Me too",
            "user": {
                "login": "octocat",
                "id": 1,
                "node_id": "MDQ6VXNlcjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octocat",
                "html_url": "https://github.com/octocat",
                "followers_url": "https://api.github.com/users/octocat/followers",
                "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                "organizations_url": "https://api.github.com/users/octocat/orgs",
                "repos_url": "https://api.github.com/users/octocat/repos",
                "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octocat/received_events",
                "type": "User",
                "site_admin": False,
            },
            "created_at": "2011-04-14T16:00:49Z",
            "updated_at": "2011-04-14T16:00:49Z",
            "issue_url": "https://api.github.com/repos/octocat/Hello-World/issues/1347",
            "author_association": "COLLABORATOR",
        }
        comment = Comment.from_dict(data)

        self.assertEqual(comment.id, 1)
        self.assertEqual(comment.node_id, "MDEyOklzc3VlQ29tbWVudDE=")
        self.assertEqual(
            comment.url,
            "https://api.github.com/repos/octocat/Hello-World/issues/comments/1",
        )
        self.assertEqual(
            comment.html_url,
            "https://github.com/octocat/Hello-World/issues/1347#issuecomment-1",
        )
        self.assertEqual(comment.body, "Me too")

        user = comment.user
        self.assertEqual(user.login, "octocat")
        self.assertEqual(user.id, 1)
        self.assertEqual(user.node_id, "MDQ6VXNlcjE=")
        self.assertEqual(
            user.avatar_url, "https://github.com/images/error/octocat_happy.gif"
        )
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/octocat")
        self.assertEqual(user.html_url, "https://github.com/octocat")
        self.assertEqual(
            user.followers_url, "https://api.github.com/users/octocat/followers"
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/octocat/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/octocat/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/octocat/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/octocat/subscriptions",
        )
        self.assertEqual(
            user.organizations_url, "https://api.github.com/users/octocat/orgs"
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/octocat/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/octocat/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/octocat/received_events",
        )
        self.assertEqual(user.type, "User")
        self.assertFalse(user.site_admin)

        self.assertEqual(
            comment.issue_url,
            "https://api.github.com/repos/octocat/Hello-World/issues/1347",
        )
        self.assertEqual(
            comment.author_association, AuthorAssociation.COLLABORATOR
        )
